/*
 * pwmh.c
 *
 *  Created on: 20 mai 2021
 *      Author: Florian
 */

/* Local includes */
#include "pwmh.h"

/* Local constants and macros */

/* Local typedefs and enums */
typedef enum
{
	cPWMH_STM_RESET, 			// Reset state
	cPWMH_STM_OFF,   			// OFF state - set point is 0
	cPWMH_STM_KICKSTART_RAMP,	// PWM is transitioned from 0 to kickstart threshold
	cPWMH_STM_KICKSTART,		// PWM is maintained at kickstart threshold during kickstart delay
	cPWMH_STM_RAMP, 			// PWM is transitioned to set point
	cPWMH_STM_SET, 				// PWM is fixed, set point is reached
} tPWMH_stm;

typedef struct
{
	tPWMH_stm stm_T;
	uint8_t kickstart_threshold_u8; 				// Kickstart duty cycle
	uint8_t dyc_u8, dyc_prev_u8;					// Actual duty cycle, previous duty cycle before ramp
	uint16_t ramp_cnt_u16, ramp_cumul_u16;			// Internal counters
	uint16_t ramp_time_mem_u16;						// Memorised ramp time for set point
	uint16_t dyc_setpoint_u8;						// Set point duty cycle
	BOOL inc_B;										// Ramp direction
	uint16_t kickstart_delay_u16;					// Memorised kickstart delay
	uint16_t kickstart_ramp_time_u16;				// Memorised kickstart ramp time
	uint8_t delta_u8;								// Diff between start duty cycle and set point
	BOOL new_setpoint_B;							// Indicates that a new set point has been set
}PWMH_Handle_t;

typedef struct
{
	TIM_HandleTypeDef *htim_T;
	uint32_t channel_u32;
}PWMH_Param_t;

/* Local variables */

static PWMH_Handle_t PWMH_Handles_TA  [cPWMH_PWM_NUMBER];
static PWMH_Param_t PWM_Params_TA [cPWMH_PWM_NUMBER] =
{
		{&htim1, TIM_CHANNEL_1},
		{&htim1, TIM_CHANNEL_2},
		{&htim1, TIM_CHANNEL_3},
		{&htim1, TIM_CHANNEL_4},
		{&htim4, TIM_CHANNEL_1},
		{&htim4, TIM_CHANNEL_2},
		{&htim4, TIM_CHANNEL_3},
		{&htim4, TIM_CHANNEL_4},
		{&htim8, TIM_CHANNEL_1},
		{&htim8, TIM_CHANNEL_2},
};

/* Local static functions */

static void SetDYC (uint8_t channel_u8, uint8_t dyc_u8)
{
	__HAL_TIM_SET_COMPARE((PWM_Params_TA[channel_u8].htim_T), (PWM_Params_TA[channel_u8].channel_u32), (uint32_t)dyc_u8);
}

static void StmUpdate (uint8_t channel_u8)
{
	/* PWM1 main state machine*/
	switch (PWMH_Handles_TA[channel_u8].stm_T)
	{
	case cPWMH_STM_RESET:
		PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_OFF;
		break;

	case cPWMH_STM_OFF:
		if (PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 > 0U) // PWM is activated
		{
			PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer

			if (PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 < PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
			{ // Kickstart is required - Start ramp at kickstart threshold
				PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_KICKSTART_RAMP;
				PWMH_Handles_TA[channel_u8].inc_B = FALSE; // Set point is lower than the kickstart threshold,
									//   thus the ramp after the kickstart will be downhill
				PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8 - PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
				if (PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16 > PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 >> 1U;
				}
				else
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].delta_u8 >> 1U;
				}
			}
			else
			{ // No kickstart required - Start ramp at zero
				PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_RAMP;
				PWMH_Handles_TA[channel_u8].dyc_prev_u8 = 0U;
				PWMH_Handles_TA[channel_u8].inc_B = TRUE; // Start is 0, thus ramp is uphill
				PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;

				if (PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 >> 1U;
				}
				else
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].delta_u8 >> 1U;
				}
			}
		}
		break;

	case cPWMH_STM_KICKSTART_RAMP:
		if (PWMH_Handles_TA[channel_u8].dyc_u8 >= PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
		{
			PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer
			PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
			PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_KICKSTART;
		}
		else
		{ // Continue Bresenham algo
			if (PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16 > PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
			{ // More delay than DYC delta : wait before incrementing

				if (PWMH_Handles_TA[channel_u8].ramp_cnt_u16 < PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16)
				{
					PWMH_Handles_TA[channel_u8].ramp_cnt_u16++;
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 += PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;

					if (PWMH_Handles_TA[channel_u8].ramp_cumul_u16 > PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16)
					{ // Time to increment the PWM
						PWMH_Handles_TA[channel_u8].ramp_cumul_u16 -= PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16;
						PWMH_Handles_TA[channel_u8].dyc_u8++;
					}
				}
			}
			else
			{ // More DYC delta than delay : increment at each pass

				while (PWMH_Handles_TA[channel_u8].ramp_cnt_u16 < PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
				{
					PWMH_Handles_TA[channel_u8].ramp_cnt_u16++;
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 += PWMH_Handles_TA[channel_u8].kickstart_ramp_time_u16;

					if (PWMH_Handles_TA[channel_u8].ramp_cumul_u16 > PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
					{ // Step is applied
						PWMH_Handles_TA[channel_u8].ramp_cumul_u16 -= PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
						PWMH_Handles_TA[channel_u8].dyc_u8 += PWMH_Handles_TA[channel_u8].ramp_cnt_u16;

						if (PWMH_Handles_TA[channel_u8].dyc_u8 > PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
						{
							PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
						}
						break;
					}
				}
			}
		}
		break;

	case cPWMH_STM_KICKSTART:
		PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
		PWMH_Handles_TA[channel_u8].ramp_cnt_u16++;

		// Check eventual set point update
		if (PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 < PWMH_Handles_TA[channel_u8].kickstart_threshold_u8)
		{
			PWMH_Handles_TA[channel_u8].inc_B = FALSE; // Set point is lower than the kickstart threshold,
								//   thus the ramp after the kickstart will be downhill
			PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8 - PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
			PWMH_Handles_TA[channel_u8].dyc_prev_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
		}
		else
		{ // No kickstart required - Start ramp at zero
			PWMH_Handles_TA[channel_u8].dyc_prev_u8 = PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
			PWMH_Handles_TA[channel_u8].inc_B = TRUE; // Start is 0, thus ramp is uphill
			PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 - PWMH_Handles_TA[channel_u8].kickstart_threshold_u8;
		}

		if (PWMH_Handles_TA[channel_u8].ramp_cnt_u16 >= PWMH_Handles_TA[channel_u8].kickstart_delay_u16)
		{
			PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer
			PWMH_Handles_TA[channel_u8].dyc_prev_u8 = PWMH_Handles_TA[channel_u8].dyc_u8; // Latch start value
			if (PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
			{
				PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 >> 1U;
			}
			else
			{
				PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].delta_u8 >> 1U;
			}
			PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_RAMP;
		}
		break;

	case cPWMH_STM_RAMP:
		if (PWMH_Handles_TA[channel_u8].dyc_u8 == PWMH_Handles_TA[channel_u8].dyc_setpoint_u8)
		{
			PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer
			PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_SET;
		}
		else
		{ // Continue Bresenham algo
			if (PWMH_Handles_TA[channel_u8].new_setpoint_B == TRUE)
			{ // If incoming new values
				PWMH_Handles_TA[channel_u8].new_setpoint_B = FALSE; // Reset flag
				PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer
				PWMH_Handles_TA[channel_u8].dyc_prev_u8 = PWMH_Handles_TA[channel_u8].dyc_u8; // Latch start value
				if (PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 >> 1U;
				}
				else
				{
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].delta_u8 >> 1U;
				}
			}

			if (PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
			{ // More delay than DYC delta : wait before incrementing

				if (PWMH_Handles_TA[channel_u8].ramp_cnt_u16 < PWMH_Handles_TA[channel_u8].ramp_time_mem_u16)
				{
					PWMH_Handles_TA[channel_u8].ramp_cnt_u16++;
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 += PWMH_Handles_TA[channel_u8].delta_u8;

					if (PWMH_Handles_TA[channel_u8].ramp_cumul_u16 > PWMH_Handles_TA[channel_u8].ramp_time_mem_u16)
					{ // Time to increment the PWM
						PWMH_Handles_TA[channel_u8].ramp_cumul_u16 -= PWMH_Handles_TA[channel_u8].ramp_time_mem_u16;

						if (PWMH_Handles_TA[channel_u8].inc_B == TRUE)
						{
							PWMH_Handles_TA[channel_u8].dyc_u8++;
						}
						else
						{
							PWMH_Handles_TA[channel_u8].dyc_u8--;
						}
					}
				}
			}
			else
			{ // More DYC delta than delay : increment at each pass

				while (PWMH_Handles_TA[channel_u8].ramp_cnt_u16 < PWMH_Handles_TA[channel_u8].delta_u8)
				{
					PWMH_Handles_TA[channel_u8].ramp_cnt_u16++;
					PWMH_Handles_TA[channel_u8].ramp_cumul_u16 += PWMH_Handles_TA[channel_u8].ramp_time_mem_u16;

					if (PWMH_Handles_TA[channel_u8].ramp_cumul_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
					{ // Step is applied
						PWMH_Handles_TA[channel_u8].ramp_cumul_u16 -= PWMH_Handles_TA[channel_u8].delta_u8;

						if (PWMH_Handles_TA[channel_u8].inc_B == TRUE)
						{
							PWMH_Handles_TA[channel_u8].dyc_u8 += PWMH_Handles_TA[channel_u8].ramp_cnt_u16;
							if (PWMH_Handles_TA[channel_u8].dyc_u8 > PWMH_Handles_TA[channel_u8].dyc_setpoint_u8)
							{
								PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
							}
						}
						else
						{
							PWMH_Handles_TA[channel_u8].dyc_u8 -= PWMH_Handles_TA[channel_u8].ramp_cnt_u16;
							if (PWMH_Handles_TA[channel_u8].dyc_u8 < PWMH_Handles_TA[channel_u8].dyc_setpoint_u8)
							{
								PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
							}
						}
						break;
					}
				}
			}
		}
		break;

	case cPWMH_STM_SET:
		if (PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 != PWMH_Handles_TA[channel_u8].dyc_u8) // PWM is changed
		{
			PWMH_Handles_TA[channel_u8].ramp_cnt_u16 = 0U; // Reset timer
			PWMH_Handles_TA[channel_u8].dyc_prev_u8 = PWMH_Handles_TA[channel_u8].dyc_u8; // Latch start value

			if (PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 > PWMH_Handles_TA[channel_u8].delta_u8)
			{
				PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 >> 1U;
			}
			else
			{
				PWMH_Handles_TA[channel_u8].ramp_cumul_u16 = PWMH_Handles_TA[channel_u8].delta_u8 >> 1U;
			}

			if (PWMH_Handles_TA[channel_u8].dyc_u8 < PWMH_Handles_TA[channel_u8].dyc_setpoint_u8)
			{
				PWMH_Handles_TA[channel_u8].inc_B = TRUE;
				PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 - PWMH_Handles_TA[channel_u8].dyc_prev_u8;
			}
			else
			{
				PWMH_Handles_TA[channel_u8].inc_B = FALSE;
				PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_prev_u8 - PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
			}

			PWMH_Handles_TA[channel_u8].stm_T = cPWMH_STM_RAMP;
		}
		break;

	default:
		break;
	}
}

/* Shared functions */

void PWMH_Init (void)
{
	uint8_t cnt_u8 = 0U;

	MX_TIM1_Init();
	MX_TIM4_Init();
	MX_TIM8_Init();

	for (cnt_u8 = 0U; cnt_u8 < cPWMH_PWM_NUMBER; cnt_u8++)
	{
		PWMH_Handles_TA[cnt_u8].stm_T = cPWMH_STM_RESET;
		PWMH_Handles_TA[cnt_u8].dyc_setpoint_u8 = 0U;
		PWMH_Handles_TA[cnt_u8].dyc_u8 = 0U;
		PWMH_Handles_TA[cnt_u8].dyc_prev_u8 = 0U;
		PWMH_Handles_TA[cnt_u8].ramp_cnt_u16 = 0U;
		PWMH_Handles_TA[cnt_u8].ramp_cumul_u16 = 0U;
		PWMH_Handles_TA[cnt_u8].ramp_time_mem_u16 = 0U;
		PWMH_Handles_TA[cnt_u8].inc_B = TRUE;
		PWMH_Handles_TA[cnt_u8].kickstart_threshold_u8 = cPWMH_KICKSTART_THRESHOLD_DEFAULT_U8;
		PWMH_Handles_TA[cnt_u8].kickstart_delay_u16 = cPWMH_KICKSTART_DELAY_DEFAULT_U16;
		PWMH_Handles_TA[cnt_u8].kickstart_ramp_time_u16 = cPWMH_KICKSTART_RAMP_TIME_DEFAULT_U16;
		PWMH_Handles_TA[cnt_u8].delta_u8 = 0U;
		PWMH_Handles_TA[cnt_u8].new_setpoint_B = FALSE;
	}
}

void PWMH_Start (void)
{
	uint8_t cnt_u8 = 0U;

	for (cnt_u8 = 0U; cnt_u8 < cPWMH_PWM_NUMBER; cnt_u8++)
	{
		HAL_TIM_PWM_Start_IT(PWM_Params_TA[cnt_u8].htim_T, PWM_Params_TA[cnt_u8].channel_u32);
	}
}

void PWMH_Stop (void)
{
	uint8_t cnt_u8 = 0U;

	for (cnt_u8 = 0U; cnt_u8 < cPWMH_PWM_NUMBER; cnt_u8++)
	{
		HAL_TIM_PWM_Stop_IT(PWM_Params_TA[cnt_u8].htim_T, PWM_Params_TA[cnt_u8].channel_u32);
	}
}

void PWMH_PeriodicFunction (void)
{
	uint8_t cnt_u8 = 0U;

	for (cnt_u8 = 0U; cnt_u8 < cPWMH_PWM_NUMBER; cnt_u8++)
	{
		StmUpdate(cnt_u8);
		SetDYC(cnt_u8, PWMH_Handles_TA[cnt_u8].dyc_u8);
	}

}

BOOL PWMH_SetRamp (uint8_t channel_u8, uint8_t dyc_u8, uint16_t ramp_time_u16)
{
	BOOL ret_B = FALSE;

	PWMH_Handles_TA[channel_u8].new_setpoint_B = TRUE;

	if (dyc_u8 != PWMH_Handles_TA[channel_u8].dyc_u8)
	{
		PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 = dyc_u8;

		if (ramp_time_u16 == 0U) // Ramp time 0 : apply DYC instantly
		{
			PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 = ramp_time_u16;
			PWMH_Handles_TA[channel_u8].dyc_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
			ret_B = TRUE;
			return ret_B;
		}

		PWMH_Handles_TA[channel_u8].ramp_time_mem_u16 = ramp_time_u16;

		if (dyc_u8 > PWMH_Handles_TA[channel_u8].dyc_u8)
		{
			PWMH_Handles_TA[channel_u8].inc_B = TRUE;
			PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_setpoint_u8 - PWMH_Handles_TA[channel_u8].dyc_u8;
		}
		else
		{
			PWMH_Handles_TA[channel_u8].inc_B = FALSE;
			PWMH_Handles_TA[channel_u8].delta_u8 = PWMH_Handles_TA[channel_u8].dyc_u8 - PWMH_Handles_TA[channel_u8].dyc_setpoint_u8;
		}
	}

	return ret_B;
}
